@@ -2,7 +2,7 @@ from collections import defaultdict
 from fastapi import APIRouter, Query, Path, Cookie
 from fastapi.encoders import jsonable_encoder
 from fastapi.requests import Request
-from fastapi.responses import RedirectResponse, JSONResponse
+from fastapi.responses import RedirectResponse, JSONResponse, Response
 from sovereign import html_templates, discovery, XDS_TEMPLATES
 from sovereign.discovery import DiscoveryTypes
 from sovereign import json_response_class
@@ -84,8 +84,7 @@ async def resources(
     except KeyError:
         ret['resources'] = []
     else:
-        if isinstance(response, dict):
-            ret['resources'] += response.get('resources') or []
+        ret['resources'] += response.resources
     return html_templates.TemplateResponse(
         name='resources.html',
         media_type='text/html',
@@ -122,11 +121,7 @@ async def resource(
         ),
         xds_type=xds_type.value
     )
-    safe_response = jsonable_encoder(response)
-    try:
-        return json_response_class(content=safe_response)
-    except TypeError:
-        return JSONResponse(content=safe_response)
+    return Response(response.rendered, media_type='application/json')
 
 
 @router.get(
@@ -149,18 +144,17 @@ async def virtual_hosts(
         ),
         xds_type='routes'
     )
-    if isinstance(response, dict):
-        route_configs = [
-            resource_
-            for resource_ in response.get('resources', [])
-            if resource_['name'] == route_configuration
-        ]
-        for route_config in route_configs:
-            for vhost in route_config['virtual_hosts']:
-                if vhost['name'] == virtual_host:
-                    safe_response = jsonable_encoder(vhost)
-                    try:
-                        return json_response_class(content=safe_response)
-                    except TypeError:
-                        return JSONResponse(content=safe_response)
-            break
+    route_configs = [
+        resource_
+        for resource_ in response.resources
+        if resource_['name'] == route_configuration
+    ]
+    for route_config in route_configs:
+        for vhost in route_config['virtual_hosts']:
+            if vhost['name'] == virtual_host:
+                safe_response = jsonable_encoder(vhost)
+                try:
+                    return json_response_class(content=safe_response)
+                except TypeError:
+                    return JSONResponse(content=safe_response)
+        break
